@@ -8,10 +8,14 @@
 #  rights to use.
 ####################################################################
 
+from typing import Optional, Dict
 import __future__
 import sys
-import numpy
+import os
 import threading
+import re
+
+import numpy
 
 is_cpu_amd_intel = False # DEPRECATION WARNING: WILL BE REMOVED IN FUTURE RELEASE
 from numexpr import interpreter, expressions, use_vml
@@ -19,22 +23,21 @@ from numexpr.utils import CacheDict
 
 # Declare a double type that does not exist in Python space
 double = numpy.double
-if sys.version_info[0] < 3:
-    int_ = int
-    long_ = int
-else:
-    int_ = numpy.int32
-    long_ = numpy.int64
+double = numpy.double
 
-typecode_to_kind = {'b': 'bool', 'i': 'int', 'l': 'long', 'f': 'float',
-                    'd': 'double', 'c': 'complex', 's': 'bytes', 'n': 'none'}
-kind_to_typecode = {'bool': 'b', 'int': 'i', 'long': 'l', 'float': 'f',
-                    'double': 'd', 'complex': 'c', 'bytes': 's', 'none': 'n'}
+int_ = numpy.int32
+long_ = numpy.int64
+
+typecode_to_kind = {'b': 'bool', 'i': 'int', 'l': 'long', 'f': 'float', 'd': 'double', 
+                    'c': 'complex', 'n': 'none', 's': 'str'}
+kind_to_typecode = {'bool': 'b', 'int': 'i', 'long': 'l', 'float': 'f', 'double': 'd',
+                    'complex': 'c', 'bytes': 's', 'str': 's', 'none': 'n'}
 type_to_typecode = {bool: 'b', int_: 'i', long_: 'l', float: 'f',
-                    double: 'd', complex: 'c', bytes: 's'}
+                    double: 'd', complex: 'c', bytes: 's', str: 's'}
 type_to_kind = expressions.type_to_kind
 kind_to_type = expressions.kind_to_type
 default_type = kind_to_type[expressions.default_kind]
+scalar_constant_kinds = list(kind_to_typecode.keys())
 
 # VML functions that are implemented in numexpr
 vml_functions = [
@@ -68,16 +71,8 @@ vml_functions = [
     "floor"
     ]
 
-# Final addtions for Python 3 (mainly for PyTables needs)
-if sys.version_info[0] > 2:
-    typecode_to_kind['s'] = 'str'
-    kind_to_typecode['str'] = 's'
-    type_to_typecode[str] = 's'
 
-scalar_constant_kinds = list(kind_to_typecode.keys())
-
-
-class ASTNode(object):
+class ASTNode():
     """Abstract Syntax Tree node.
 
     Members:
@@ -91,9 +86,7 @@ class ASTNode(object):
     """
     cmpnames = ['astType', 'astKind', 'value', 'children']
 
-    def __init__(self, astType='generic', astKind='unknown',
-                 value=None, children=()):
-        object.__init__(self)
+    def __init__(self, astType='generic', astKind='unknown', value=None, children=()):
         self.astType = astType
         self.astKind = astKind
         self.value = value
@@ -122,7 +115,7 @@ class ASTNode(object):
                 return numpy.array(self.value) < numpy.array(other.value)
             return self.astKind < other.astKind
         else:
-            raise TypeError( 'Sorting not implemented for astType: %s'%self.astType )
+            raise TypeError('Sorting not implemented for astType: %s'%self.astType)
 
     def __hash__(self):
         if self.astType == 'alias':
@@ -186,9 +179,9 @@ def sigPerms(s):
 
 
 def typeCompileAst(ast):
-    """Assign appropiate types to each node in the AST.
+    """Assign appropriate types to each node in the AST.
 
-    Will convert opcodes and functions to appropiate upcast version,
+    Will convert opcodes and functions to appropriate upcast version,
     and add "cast" ops if needed.
     """
     children = list(ast.children)
@@ -228,7 +221,7 @@ def typeCompileAst(ast):
                    [typeCompileAst(c) for c in children])
 
 
-class Register(object):
+class Register():
     """Abstraction for a register in the VM.
 
     Members:
@@ -270,9 +263,24 @@ class Immediate(Register):
         return 'Immediate(%d)' % (self.node.value,)
 
 
-def stringToExpression(s, types, context):
+_flow_pat = r'[\;\[\:]'
+_dunder_pat = r'(^|[^\w])__[\w]+__($|[^\w])'
+_attr_pat = r'\.\b(?!(real|imag|(\d*[eE]?[+-]?\d+)|\d*j)\b)'
+_blacklist_re = re.compile(f'{_flow_pat}|{_dunder_pat}|{_attr_pat}')
+
+def stringToExpression(s, types, context, sanitize: bool=True):
     """Given a string, convert it to a tree of ExpressionNode's.
     """
+    # sanitize the string for obvious attack vectors that NumExpr cannot 
+    # parse into its homebrew AST. This is to protect the call to `eval` below.
+    # We forbid `;`, `:`. `[` and `__`, and attribute access via '.'.
+    # We cannot ban `.real` or `.imag` however...
+    # We also cannot ban `.\d*j`, where `\d*` is some digits (or none), e.g. 1.5j, 1.j
+    if sanitize:
+        no_whitespace = re.sub(r'\s+', '', s)
+        if _blacklist_re.search(no_whitespace) is not None:
+            raise ValueError(f'Expression {s} has forbidden control characters.')
+    
     old_ctx = expressions._context.get_current_context()
     try:
         expressions._context.set_new_context(context)
@@ -295,8 +303,10 @@ def stringToExpression(s, types, context):
                 t = types.get(name, default_type)
                 names[name] = expressions.VariableNode(name, type_to_kind[t])
         names.update(expressions.functions)
+
         # now build the expression
         ex = eval(c, names)
+        
         if expressions.isConstant(ex):
             ex = expressions.ConstantNode(ex, expressions.getKind(ex))
         elif not isinstance(ex, expressions.ExpressionNode):
@@ -312,7 +322,8 @@ def isReduction(ast):
 
 
 def getInputOrder(ast, input_order=None):
-    """Derive the input order of the variables in an expression.
+    """
+    Derive the input order of the variables in an expression.
     """
     variables = {}
     for a in ast.allOf('variable'):
@@ -337,20 +348,22 @@ def convertConstantToKind(x, kind):
     # Exception for 'float' types that will return the NumPy float32 type
     if kind == 'float':
         return numpy.float32(x)
-    elif sys.version_info[0] >= 3 and isinstance(x,str):
+    elif isinstance(x,str):
         return x.encode('ascii')
     return kind_to_type[kind](x)
 
 
 def getConstants(ast):
-    '''
+    """
     RAM: implemented magic method __lt__ for ASTNode to fix issues
     #88 and #209. The following test code works now, as does the test suite.
-    import numexpr as ne
-    a = 1 + 3j; b = 5.0
-    ne.evaluate( 'a*2 + 15j - b' )
-    '''
-    constants_order = sorted( ast.allOf('constant') )
+
+        import numexpr as ne
+        a = 1 + 3j; b = 5.0
+        ne.evaluate('a*2 + 15j - b')
+    """
+    constant_registers = set([node.reg for node in ast.allOf("constant")]) 
+    constants_order = sorted([r.node for r in constant_registers])
     constants = [convertConstantToKind(a.value, a.astKind)
                  for a in constants_order]
     return constants_order, constants
@@ -366,7 +379,8 @@ def sortNodesByOrder(nodes, order):
 
 
 def assignLeafRegisters(inodes, registerMaker):
-    """Assign new registers to each of the leaf nodes.
+    """
+    Assign new registers to each of the leaf nodes.
     """
     leafRegisters = {}
     for node in inodes:
@@ -378,14 +392,16 @@ def assignLeafRegisters(inodes, registerMaker):
 
 
 def assignBranchRegisters(inodes, registerMaker):
-    """Assign temporary registers to each of the branch nodes.
+    """
+    Assign temporary registers to each of the branch nodes.
     """
     for node in inodes:
         node.reg = registerMaker(node, temporary=True)
 
 
 def collapseDuplicateSubtrees(ast):
-    """Common subexpression elimination.
+    """
+    Common subexpression elimination.
     """
     seen = {}
     aliases = []
@@ -407,8 +423,8 @@ def collapseDuplicateSubtrees(ast):
 
 
 def optimizeTemporariesAllocation(ast):
-    """Attempt to minimize the number of temporaries needed, by
-    reusing old ones.
+    """
+    Attempt to minimize the number of temporaries needed, by reusing old ones.
     """
     nodes = [n for n in ast.postorderWalk() if n.reg.temporary]
     users_of = dict((n.reg, set()) for n in nodes)
@@ -440,7 +456,8 @@ def optimizeTemporariesAllocation(ast):
 
 
 def setOrderedRegisterNumbers(order, start):
-    """Given an order of nodes, assign register numbers.
+    """
+    Given an order of nodes, assign register numbers.
     """
     for i, node in enumerate(order):
         node.reg.n = start + i
@@ -448,7 +465,8 @@ def setOrderedRegisterNumbers(order, start):
 
 
 def setRegisterNumbersForTemporaries(ast, start):
-    """Assign register numbers for temporary registers, keeping track of
+    """
+    Assign register numbers for temporary registers, keeping track of
     aliases and handling immediate operands.
     """
     seen = 0
@@ -472,7 +490,8 @@ def setRegisterNumbersForTemporaries(ast, start):
 
 
 def convertASTtoThreeAddrForm(ast):
-    """Convert an AST to a three address form.
+    """
+    Convert an AST to a three address form.
 
     Three address form is (op, reg1, reg2, reg3), where reg1 is the
     destination of the result of the instruction.
@@ -485,7 +504,8 @@ def convertASTtoThreeAddrForm(ast):
 
 
 def compileThreeAddrForm(program):
-    """Given a three address form of the program, compile it a string that
+    """
+    Given a three address form of the program, compile it a string that
     the VM understands.
     """
 
@@ -495,12 +515,7 @@ def compileThreeAddrForm(program):
         elif reg.n < 0:
             raise ValueError("negative value for register number %s" % reg.n)
         else:
-            if sys.version_info[0] < 3:
-                return chr(reg.n)
-            else:
-                # int.to_bytes is not available in Python < 3.2
-                #return reg.n.to_bytes(1, sys.byteorder)
-                return bytes([reg.n])
+            return bytes([reg.n])
 
     def quadrupleToString(opcode, store, a1=None, a2=None):
         cop = chr(interpreter.opcodes[opcode]).encode('ascii')
@@ -532,7 +547,7 @@ context_info = [
 ]
 
 
-def getContext(kwargs, frame_depth=1):
+def getContext(kwargs, _frame_depth=1):
     d = kwargs.copy()
     context = {}
     for name, allowed, default in context_info:
@@ -541,24 +556,25 @@ def getContext(kwargs, frame_depth=1):
             context[name] = value
         else:
             raise ValueError("'%s' must be one of %s" % (name, allowed))
-
+    
     if d:
         raise ValueError("Unknown keyword argument '%s'" % d.popitem()[0])
     if context['truediv'] == 'auto':
-        caller_globals = sys._getframe(frame_depth + 1).f_globals
+        caller_globals = sys._getframe(_frame_depth + 1).f_globals
         context['truediv'] = caller_globals.get('division', None) == __future__.division
 
     return context
 
 
-def precompile(ex, signature=(), context={}):
-    """Compile the expression to an intermediate form.
+def precompile(ex, signature=(), context={}, sanitize: bool=True):
+    """
+    Compile the expression to an intermediate form.
     """
     types = dict(signature)
     input_order = [name for (name, type_) in signature]
 
     if isinstance(ex, str):
-        ex = stringToExpression(ex, types, context)
+        ex = stringToExpression(ex, types, context, sanitize)
 
     # the AST is like the expression, but the node objects don't have
     # any odd interpretations
@@ -604,7 +620,7 @@ def precompile(ex, signature=(), context={}):
     return threeAddrProgram, signature, tempsig, constants, input_names
 
 
-def NumExpr(ex, signature=(), **kwargs):
+def NumExpr(ex,  signature=(), sanitize: bool=True, **kwargs):
     """
     Compile an expression built using E.<variable> variables to a function.
 
@@ -615,15 +631,13 @@ def NumExpr(ex, signature=(), **kwargs):
 
     Returns a `NumExpr` object containing the compiled function.
     """
-    # NumExpr can be called either directly by the end-user, in which case
-    # kwargs need to be sanitized by getContext, or by evaluate,
-    # in which case kwargs are in already sanitized.
-    # In that case frame_depth is wrong (it should be 2) but it doesn't matter
+
+    # In that case _frame_depth is wrong (it should be 2) but it doesn't matter
     # since it will not be used (because truediv='auto' has already been
     # translated to either True or False).
-
-    context = getContext(kwargs, frame_depth=1)
-    threeAddrProgram, inputsig, tempsig, constants, input_names = precompile(ex, signature, context)
+    _frame_depth = 1
+    context = getContext(kwargs, _frame_depth=_frame_depth)
+    threeAddrProgram, inputsig, tempsig, constants, input_names = precompile(ex, signature, context, sanitize=sanitize)
     program = compileThreeAddrForm(threeAddrProgram)
     return interpreter.NumExpr(inputsig.encode('ascii'),
                                tempsig.encode('ascii'),
@@ -640,21 +654,20 @@ def disassemble(nex):
     r_constants = 1 + len(nex.signature)
     r_temps = r_constants + len(nex.constants)
 
+    def parseOp(op):
+        name, sig = [*op.rsplit(b'_', 1), ''][:2]
+        return name, sig 
+
     def getArg(pc, offset):
-        if sys.version_info[0] < 3:
-            arg = ord(nex.program[pc + offset])
-            op = rev_opcodes.get(ord(nex.program[pc]))
-        else:
-            arg = nex.program[pc + offset]
-            op = rev_opcodes.get(nex.program[pc])
+        arg = nex.program[pc + (offset if offset < 4 else offset+1)]
+        _, sig = parseOp(rev_opcodes.get(nex.program[pc]))
         try:
-            code = op.split(b'_')[1][offset - 1]
+            code = sig[offset - 1]
         except IndexError:
             return None
-        if sys.version_info[0] > 2:
-            # int.to_bytes is not available in Python < 3.2
-            #code = code.to_bytes(1, sys.byteorder)
-            code = bytes([code])
+
+        code = bytes([code])
+
         if arg == 255:
             return None
         if code != b'n':
@@ -671,14 +684,14 @@ def disassemble(nex):
 
     source = []
     for pc in range(0, len(nex.program), 4):
-        if sys.version_info[0] < 3:
-            op = rev_opcodes.get(ord(nex.program[pc]))
-        else:
-            op = rev_opcodes.get(nex.program[pc])
-        dest = getArg(pc, 1)
-        arg1 = getArg(pc, 2)
-        arg2 = getArg(pc, 3)
-        source.append((op, dest, arg1, arg2))
+        op = rev_opcodes.get(nex.program[pc])
+        _, sig = parseOp(op)
+        parsed = [op]
+        for i in range(len(sig)):
+            parsed.append(getArg(pc, 1 + i))
+        while len(parsed) < 4:
+            parsed.append(None)
+        source.append(parsed)
     return source
 
 
@@ -700,11 +713,13 @@ def getType(a):
         return complex
     if kind == 'S':
         return bytes
+    if kind == 'U':
+        raise ValueError('NumExpr 2 does not support Unicode as a dtype.')
     raise ValueError("unknown type %s" % a.dtype.name)
 
 
-def getExprNames(text, context):
-    ex = stringToExpression(text, {}, context)
+def getExprNames(text, context, sanitize: bool=True):
+    ex = stringToExpression(text, {}, context, sanitize)
     ast = expressionToAST(ex)
     input_order = getInputOrder(ast, None)
     #try to figure out if vml operations are used by expression
@@ -721,18 +736,24 @@ def getExprNames(text, context):
     return [a.value for a in input_order], ex_uses_vml
 
 
-def getArguments(names, local_dict=None, global_dict=None):
-    """Get the arguments based on the names."""
-    call_frame = sys._getframe(2)
+def getArguments(names, local_dict=None, global_dict=None, _frame_depth: int=2):
+    """
+    Get the arguments based on the names.
+    """
+    call_frame = sys._getframe(_frame_depth)
 
-    local_dict_as_arg = local_dict is not None
-
-    if not local_dict_as_arg:
+    clear_local_dict = False
+    if local_dict is None:
         local_dict = call_frame.f_locals
-
+        clear_local_dict = True
     try:
+        frame_globals = call_frame.f_globals
         if global_dict is None:
-            global_dict = call_frame.f_globals
+            global_dict = frame_globals
+
+        # If `call_frame` is the top frame of the interpreter we can't clear its 
+        # `local_dict`, because it is actually the `global_dict`.
+        clear_local_dict = clear_local_dict and not frame_globals is local_dict
 
         arguments = []
         for name in names:
@@ -745,7 +766,7 @@ def getArguments(names, local_dict=None, global_dict=None):
         # If we generated local_dict via an explicit reference to f_locals,
         # clear the dict to prevent creating extra ref counts in the caller's scope
         # See https://github.com/pydata/numexpr/issues/310
-        if not local_dict_as_arg:
+        if clear_local_dict:
             local_dict.clear()
 
     return arguments
@@ -755,34 +776,46 @@ def getArguments(names, local_dict=None, global_dict=None):
 _names_cache = CacheDict(256)
 _numexpr_cache = CacheDict(256)
 _numexpr_last = {}
-
 evaluate_lock = threading.Lock()
 
-def evaluate(ex, local_dict=None, global_dict=None,
-             out=None, order='K', casting='safe', **kwargs):
-    """Evaluate a simple array expression element-wise, using the new iterator.
-
-    ex is a string forming an expression, like "2*a+3*b". The values for "a"
-    and "b" will by default be taken from the calling function's frame
-    (through use of sys._getframe()). Alternatively, they can be specifed
-    using the 'local_dict' or 'global_dict' arguments.
+# MAYBE: decorate this function to add attributes instead of having the 
+# _numexpr_last dictionary?
+def validate(ex: str, 
+             local_dict: Optional[Dict] = None, 
+             global_dict: Optional[Dict] = None,
+             out: numpy.ndarray = None, 
+             order: str = 'K', 
+             casting: str = 'safe', 
+             _frame_depth: int = 2,
+             sanitize: Optional[bool] = None,
+             **kwargs) -> Optional[Exception]:
+    r"""
+    Validate a NumExpr expression with the given `local_dict` or `locals()`.
+    Returns `None` on success and the Exception object if one occurs. Note that 
+    you can proceed directly to call `re_evaluate()` if you use `validate()`
+    to sanitize your expressions and variables in advance.
 
     Parameters
     ----------
+    ex: str
+        a string forming an expression, like "2*a+3*b". The values for "a"
+        and "b" will by default be taken from the calling function's frame
+        (through use of sys._getframe()). Alternatively, they can be specified
+        using the 'local_dict' or 'global_dict' arguments.
 
-    local_dict : dictionary, optional
+    local_dict: dictionary, optional
         A dictionary that replaces the local operands in current frame.
 
-    global_dict : dictionary, optional
+    global_dict: dictionary, optional
         A dictionary that replaces the global operands in current frame.
 
-    out : NumPy array, optional
+    out: NumPy array, optional
         An existing array where the outcome is going to be stored.  Care is
         required so that this array has the same shape and type than the
         actual outcome of the computation.  Useful for avoiding unnecessary
         new array allocations.
 
-    order : {'C', 'F', 'A', or 'K'}, optional
+    order: {'C', 'F', 'A', or 'K'}, optional
         Controls the iteration order for operands. 'C' means C order, 'F'
         means Fortran order, 'A' means 'F' order if all the arrays are
         Fortran contiguous, 'C' order otherwise, and 'K' means as close to
@@ -790,7 +823,7 @@ def evaluate(ex, local_dict=None, global_dict=None,
         efficient computations, typically 'K'eep order (the default) is
         desired.
 
-    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
+    casting: {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
         Controls what kind of data casting may occur when making a copy or
         buffering.  Setting this to 'unsafe' is not recommended, as it can
         adversely affect accumulations.
@@ -801,37 +834,152 @@ def evaluate(ex, local_dict=None, global_dict=None,
           * 'same_kind' means only safe casts or casts within a kind,
             like float64 to float32, are allowed.
           * 'unsafe' means any data conversions may be done.
+
+    sanitize: Optional[bool]
+        Both `validate` and by extension `evaluate` call `eval(ex)`, which is 
+        potentially dangerous on unsanitized inputs. As such, NumExpr by default 
+        performs simple sanitization, banning the character ':;[', the 
+        dunder '__[\w+]__', and attribute access to all but '.real' and '.imag'.
+        
+        Using `None` defaults to `True` unless the environment variable 
+        `NUMEXPR_SANITIZE=0` is set, in which case the default is `False`. 
+        Nominally this can be set via `os.environ` before `import numexpr`.
+
+    _frame_depth: int
+        The calling frame depth. Unless you are a NumExpr developer you should 
+        not set this value.
+
+    Note
+    ----
+    
     """
     global _numexpr_last
-    if not isinstance(ex, str):
-        raise ValueError("must specify expression as a string")
-    # Get the names for this expression
-    context = getContext(kwargs, frame_depth=1)
-    expr_key = (ex, tuple(sorted(context.items())))
-    if expr_key not in _names_cache:
-        _names_cache[expr_key] = getExprNames(ex, context)
-    names, ex_uses_vml = _names_cache[expr_key]
-    arguments = getArguments(names, local_dict, global_dict)
 
-    # Create a signature
-    signature = [(name, getType(arg)) for (name, arg) in
-                 zip(names, arguments)]
-
-    # Look up numexpr if possible.
-    numexpr_key = expr_key + (tuple(signature),)
     try:
-        compiled_ex = _numexpr_cache[numexpr_key]
-    except KeyError:
-        compiled_ex = _numexpr_cache[numexpr_key] = NumExpr(ex, signature, **context)
-    kwargs = {'out': out, 'order': order, 'casting': casting,
-              'ex_uses_vml': ex_uses_vml}
-    _numexpr_last = dict(ex=compiled_ex, argnames=names, kwargs=kwargs)
-    with evaluate_lock:
-        return compiled_ex(*arguments, **kwargs)
+        
+        if not isinstance(ex, str):
+            raise ValueError("must specify expression as a string")
+        
+        if sanitize is None:
+            if 'NUMEXPR_SANITIZE' in os.environ:
+                sanitize = bool(int(os.environ['NUMEXPR_SANITIZE']))
+            else:
+                sanitize = True
 
+        # Get the names for this expression
+        context = getContext(kwargs)
+        expr_key = (ex, tuple(sorted(context.items())))
+        if expr_key not in _names_cache:
+            _names_cache[expr_key] = getExprNames(ex, context, sanitize=sanitize)
+        names, ex_uses_vml = _names_cache[expr_key]
+        arguments = getArguments(names, local_dict, global_dict, _frame_depth=_frame_depth)
 
-def re_evaluate(local_dict=None):
-    """Re-evaluate the previous executed array expression without any check.
+        # Create a signature
+        signature = [(name, getType(arg)) for (name, arg) in
+                    zip(names, arguments)]
+
+        # Look up numexpr if possible.
+        numexpr_key = expr_key + (tuple(signature),)
+        try:
+            compiled_ex = _numexpr_cache[numexpr_key]
+        except KeyError:
+            compiled_ex = _numexpr_cache[numexpr_key] = NumExpr(ex, signature, sanitize=sanitize, **context)
+        kwargs = {'out': out, 'order': order, 'casting': casting,
+                'ex_uses_vml': ex_uses_vml}
+        _numexpr_last = dict(ex=compiled_ex, argnames=names, kwargs=kwargs)
+    except Exception as e:
+        return e
+    return None
+
+def evaluate(ex: str, 
+             local_dict: Optional[Dict] = None, 
+             global_dict: Optional[Dict] = None,
+             out: numpy.ndarray = None, 
+             order: str = 'K', 
+             casting: str = 'safe', 
+             sanitize: Optional[bool] = None,
+             _frame_depth: int = 3,
+             **kwargs) -> numpy.ndarray:
+    r"""
+    Evaluate a simple array expression element-wise using the virtual machine.
+
+    Parameters
+    ----------
+    ex: str
+        a string forming an expression, like "2*a+3*b". The values for "a"
+        and "b" will by default be taken from the calling function's frame
+        (through use of sys._getframe()). Alternatively, they can be specified
+        using the 'local_dict' or 'global_dict' arguments.
+
+    local_dict: dictionary, optional
+        A dictionary that replaces the local operands in current frame.
+
+    global_dict: dictionary, optional
+        A dictionary that replaces the global operands in current frame.
+
+    out: NumPy array, optional
+        An existing array where the outcome is going to be stored.  Care is
+        required so that this array has the same shape and type than the
+        actual outcome of the computation.  Useful for avoiding unnecessary
+        new array allocations.
+
+    order: {'C', 'F', 'A', or 'K'}, optional
+        Controls the iteration order for operands. 'C' means C order, 'F'
+        means Fortran order, 'A' means 'F' order if all the arrays are
+        Fortran contiguous, 'C' order otherwise, and 'K' means as close to
+        the order the array elements appear in memory as possible.  For
+        efficient computations, typically 'K'eep order (the default) is
+        desired.
+
+    casting: {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
+        Controls what kind of data casting may occur when making a copy or
+        buffering.  Setting this to 'unsafe' is not recommended, as it can
+        adversely affect accumulations.
+
+          * 'no' means the data types should not be cast at all.
+          * 'equiv' means only byte-order changes are allowed.
+          * 'safe' means only casts which can preserve values are allowed.
+          * 'same_kind' means only safe casts or casts within a kind,
+            like float64 to float32, are allowed.
+          * 'unsafe' means any data conversions may be done.
+
+    sanitize: bool
+        Both `validate` and by extension `evaluate` call `eval(ex)`, which is 
+        potentially dangerous on unsanitized inputs. As such, NumExpr by default 
+        performs simple sanitization, banning the character ':;[', the 
+        dunder '__[\w+]__', and attribute access to all but '.real' and '.imag'.
+
+        Using `None` defaults to `True` unless the environment variable 
+        `NUMEXPR_SANITIZE=0` is set, in which case the default is `False`. 
+        Nominally this can be set via `os.environ` before `import numexpr`.
+
+    _frame_depth: int
+        The calling frame depth. Unless you are a NumExpr developer you should 
+        not set this value.
+
+    Note
+    ----
+    Both `validate` and by extension `evaluate` call `eval(ex)`, which is 
+    potentially dangerous on unsanitized inputs. As such, NumExpr does some 
+    sanitization, banning the character ':;[', the dunder '__', and attribute
+    access to all but '.r' for real and '.i' for imag access to complex numbers.
+    """
+    # We could avoid code duplication if we called validate and then re_evaluate 
+    # here, but they we have difficulties with the `sys.getframe(2)` call in
+    # `getArguments`
+    e = validate(ex, local_dict=local_dict, global_dict=global_dict, 
+                 out=out, order=order, casting=casting, 
+                 _frame_depth=_frame_depth, sanitize=sanitize, **kwargs)
+    if e is None:
+        return re_evaluate(local_dict=local_dict, global_dict=global_dict, _frame_depth=_frame_depth)
+    else:
+        raise e
+    
+def re_evaluate(local_dict: Optional[Dict] = None, 
+                global_dict: Optional[Dict] = None,
+                _frame_depth: int=2) -> numpy.ndarray:
+    """
+    Re-evaluate the previous executed array expression without any check.
 
     This is meant for accelerating loops that are re-evaluating the same
     expression repeatedly without changing anything else than the operands.
@@ -839,17 +987,20 @@ def re_evaluate(local_dict=None):
 
     Parameters
     ----------
-
-    local_dict : dictionary, optional
+    local_dict: dictionary, optional
         A dictionary that replaces the local operands in current frame.
-
+    _frame_depth: int
+        The calling frame depth. Unless you are a NumExpr developer you should 
+        not set this value.
     """
+    global _numexpr_last
+
     try:
         compiled_ex = _numexpr_last['ex']
     except KeyError:
-        raise RuntimeError("not a previous evaluate() execution found")
+        raise RuntimeError("A previous evaluate() execution was not found, please call `validate` or `evaluate` once before `re_evaluate`")
     argnames = _numexpr_last['argnames']
-    args = getArguments(argnames, local_dict)
+    args = getArguments(argnames, local_dict, global_dict, _frame_depth=_frame_depth)
     kwargs = _numexpr_last['kwargs']
     with evaluate_lock:
         return compiled_ex(*args, **kwargs)
