@@ -1,3 +1,4 @@
+
 ###################################################################
 #  Numexpr - Fast numerical array expression evaluator for NumPy.
 #
@@ -19,30 +20,27 @@ import subprocess
 import numpy as np
 from numpy import (
     array, arange, empty, zeros, int32, int64, uint16, complex_, float64, rec,
-    copy, ones_like, where, alltrue, linspace,
+    copy, ones_like, where, all as alltrue, linspace,
     sum, prod, sqrt, fmod, floor, ceil,
     sin, cos, tan, arcsin, arccos, arctan, arctan2,
     sinh, cosh, tanh, arcsinh, arccosh, arctanh,
     log, log1p, log10, exp, expm1, conj)
+import numpy
 from numpy.testing import (assert_equal, assert_array_equal,
                            assert_array_almost_equal, assert_allclose)
 from numpy import shape, allclose, array_equal, ravel, isnan, isinf
 
 import numexpr
-from numexpr import E, NumExpr, evaluate, re_evaluate, disassemble, use_vml
+from numexpr import E, NumExpr, evaluate, re_evaluate, validate, disassemble, use_vml
+from numexpr.expressions import ConstantNode
 
 import unittest
 
 TestCase = unittest.TestCase
 
 double = np.double
-if sys.version_info[0] >= 3:
-    long = int
+long = int
 
-# Recommended minimum versions
-from distutils.version import LooseVersion
-minimum_numpy_version = LooseVersion('1.7.0')
-present_numpy_version = LooseVersion(np.__version__)
 
 class test_numexpr(TestCase):
     """Testing with 1 thread"""
@@ -372,23 +370,42 @@ class test_evaluate(TestCase):
         assert_array_equal(x, array([86., 124., 168.]))
 
     def test_re_evaluate_dict(self):
+        a1 = array([1., 2., 3.])
+        b1 = array([4., 5., 6.])
+        c1 = array([7., 8., 9.])
+        local_dict={'a': a1, 'b': b1, 'c': c1}
+        x = evaluate("2*a + 3*b*c", local_dict=local_dict)
+        x = re_evaluate(local_dict=local_dict)
+        assert_array_equal(x, array([86., 124., 168.]))
+
+    def test_validate(self):
         a = array([1., 2., 3.])
         b = array([4., 5., 6.])
         c = array([7., 8., 9.])
-        x = evaluate("2*a + 3*b*c", local_dict={'a': a, 'b': b, 'c': c})
+        retval = validate("2*a + 3*b*c")
+        assert(retval is None)
         x = re_evaluate()
         assert_array_equal(x, array([86., 124., 168.]))
 
-    # Test for issue #37
-    if sys.version_info[0] < 3:
-        # In python 3 '/' perforns true division, not integer division.
-        # Integer division '//' is still not suppoerted by numexpr
-        def test_zero_div(self):
-            x = arange(100, dtype='i4')
-            y = evaluate("1/x")
-            x2 = zeros(100, dtype='i4')
-            x2[1] = 1
-            assert_array_equal(x2, y)
+    def test_validate_missing_var(self):
+        a = array([1., 2., 3.])
+        b = array([4., 5., 6.])
+        retval = validate("2*a + 3*b*c")
+        assert(isinstance(retval, KeyError))
+
+    def test_validate_syntax(self):
+        retval = validate("2+")
+        assert(isinstance(retval, SyntaxError))
+
+    def test_validate_dict(self):
+        a1 = array([1., 2., 3.])
+        b1 = array([4., 5., 6.])
+        c1 = array([7., 8., 9.])
+        local_dict={'a': a1, 'b': b1, 'c': c1}
+        retval = validate("2*a + 3*b*c", local_dict=local_dict)
+        assert(retval is None)
+        x = re_evaluate(local_dict=local_dict)
+        assert_array_equal(x, array([86., 124., 168.]))
 
     # Test for issue #22
     def test_true_div(self):
@@ -487,11 +504,152 @@ class test_evaluate(TestCase):
         a = arange(3)
         try:
             evaluate("a < [0, 0, 0]")
-        except TypeError:
+        except (ValueError, TypeError):
             pass
         else:
             self.fail()
 
+    def test_sanitize(self):
+        with _environment('NUMEXPR_SANITIZE', '1'):
+            # Forbid dunder
+            try:
+                evaluate('__builtins__')
+            except ValueError:
+                pass
+            else:
+                self.fail()
+
+            # Forbid colon for lambda funcs
+            try: 
+                evaluate('lambda x: x')
+            except ValueError:
+                pass
+            else:
+                self.fail()
+
+            # Forbid indexing
+            try:
+                evaluate('locals()["evaluate"]')
+            except ValueError:
+                pass
+            else:
+                self.fail()
+
+            # Forbid semicolon
+            try:
+                evaluate('import os;')
+            except ValueError:
+                pass
+            else:
+                self.fail()
+
+            # Attribute access with spaces
+            try:
+                evaluate('os. cpu_count()')
+            except ValueError:
+                pass
+            else:
+                self.fail()
+
+            # Attribute access with funny unicode characters that eval translates
+            # into ASCII.
+            try:
+                evaluate("(3+1).ᵇit_length()")
+            except ValueError:
+                pass
+            else:
+                self.fail()
+
+            # Pass decimal points including scientific notation
+            a = 3.0
+            evaluate('a*2.e-5')
+            evaluate('a*2.e+5')
+            evaluate('a*2e-5')
+            evaluate('a*2e+5')
+            evaluate('a*2E-5')
+            evaluate('a*2.0e5')
+            evaluate('a*2.2e5')
+            evaluate('2.+a')
+
+            # pass .real and .imag
+            c = 2.5 + 1.5j
+            evaluate('c.real')
+            evaluate('c.imag')
+
+        
+    def test_no_sanitize(self):
+        try: # Errors on compile() after eval()
+            evaluate('import os;', sanitize=False)
+        except SyntaxError:
+            pass
+        else:
+            self.fail()
+
+        with _environment('NUMEXPR_SANITIZE', '0'):
+            try: # Errors on compile() after eval()
+                evaluate('import os;', sanitize=None)
+            except SyntaxError:
+                pass
+            else:
+                self.fail()
+
+    def test_disassemble(self):
+        assert_equal(disassemble(NumExpr(
+            "where(m, a, -1)", [('m', bool), ('a', float)])),
+            [[b'where_fbff', b'r0', b'r1[m]', b'r2[a]', b'c3[-1.0]'], 
+             [b'noop', None, None, None]])
+
+    def test_constant_deduplication(self):
+        assert_equal(NumExpr("(a + 1)*(a - 1)", [('a', np.int32)]).constants, (1,))
+
+    def test_nan_constant(self):
+        assert_equal(str(ConstantNode(float("nan")).value), 'nan')
+
+        # check de-duplication works for nan
+        _nan = ConstantNode(float("nan"))
+        expr = (E.a + _nan)*(E.b + _nan)
+        assert_equal(NumExpr(expr, [('a', double), ('b', double)]).constants, (float("nan"),))
+
+
+    def test_f32_constant(self):
+        assert_equal(ConstantNode(numpy.float32(1)).astKind, "float")
+        assert_equal(ConstantNode(numpy.float32("nan")).astKind, "float")
+        assert_equal(ConstantNode(numpy.float32(3)).value.dtype, numpy.dtype("float32"))
+        assert_array_equal(NumExpr(ConstantNode(numpy.float32(1))).run(), 
+                           numpy.array(1, dtype="float32"))
+
+    def test_unaligned_singleton(self):
+        # Test for issue #397 whether singletons outputs assigned to consts must be 
+        # aligned or not.
+        a = np.empty(5, dtype=np.uint8)[1:].view(np.int32)
+        evaluate('3', out=a)
+        assert_equal(a, 3)
+
+    def test_negative_mod(self):
+        # Test for issue #413, modulus of negative integers. C modulus is 
+        # actually remainder op, and hence different from Python modulus.
+        a = np.array([-500, -135, 0, 0, 135, 500], dtype=np.int32)
+        n = np.array([-360, -360, -360, 360, 360, 360], dtype=np.int32)
+        out_i = evaluate('a % n')
+        assert_equal(out_i, np.mod(a, n))
+
+        b = a.astype(np.int64)
+        m = n.astype(np.int64)
+        out_l = evaluate('b % m')
+        assert_equal(out_l, np.mod(b, m))
+
+    def test_negative_power_scalar(self):
+        # Test for issue #428, where the power is negative and the base is an
+        # integer. This was running afoul in the precomputation in `expressions.py:pow_op()`
+        base = np.array([-2, -1, 0, 1, 2, 3], dtype=np.int32)
+        out_i = evaluate('base ** -1.0')
+        assert_equal(out_i, np.power(base, -1.0))
+
+        base = np.array([-2, -1, 0, 1, 2, 3], dtype=np.int64)
+        out_l = evaluate('base ** -1.0')
+        assert_equal(out_l, np.power(base, -1.0))
+
+
     def test_ex_uses_vml(self):
         vml_funcs = [ "sin", "cos", "tan", "arcsin", "arccos", "arctan",
                       "sinh", "cosh", "tanh", "arcsinh", "arccosh", "arctanh",
@@ -604,7 +762,7 @@ def equal(a, b, exact):
         return (shape(a) == shape(b)) and alltrue(ravel(a) == ravel(b), axis=0)
     else:
         if hasattr(a, 'dtype') and a.dtype == 'f4':
-            atol = 1e-5  # Relax precission for special opcodes, like fmod
+            atol = 1e-5  # Relax precision for special opcodes, like fmod
         else:
             atol = 1e-8
         return (shape(a) == shape(b) and
@@ -727,13 +885,6 @@ class test_int64(TestCase):
 
 
 class test_int32_int64(TestCase):
-    if sys.version_info[0] < 2:
-        # no long literals in python 3
-        def test_small_long(self):
-            # Small longs should not be downgraded to ints.
-            res = evaluate('42L')
-            assert_array_equal(res, 42)
-            self.assertEqual(res.dtype.name, 'int64')
 
     def test_small_int(self):
         # Small ints (32-bit ones) should not be promoted to longs.
@@ -811,14 +962,9 @@ class test_strings(TestCase):
         str_list = [
             b'\0\0\0', b'\0\0foo\0', b'\0\0foo\0b', b'\0\0foo\0b\0',
             b'foo\0', b'foo\0b', b'foo\0b\0', b'foo\0bar\0baz\0\0']
-        min_tobytes_version = LooseVersion('1.9.0')
         for s in str_list:
             r = evaluate('s')
-            if present_numpy_version >= min_tobytes_version:
-                self.assertEqual(s, r.tobytes())  # check *all* stored data
-            else:
-                # ndarray.tostring() is deprecated as of NumPy 1.19
-                self.assertEqual(s, r.tostring())  # check *all* stored data
+            self.assertEqual(s, r.tobytes())  # check *all* stored data
 
     def test_compare_copy(self):
         sarr = self.str_array1
@@ -991,11 +1137,17 @@ class test_threading_config(TestCase):
         with _environment('OMP_NUM_THREADS', '5'):
             # NUMEXPR_NUM_THREADS has priority
             with _environment('NUMEXPR_NUM_THREADS', '3'):
-                self.assertEqual(3, numexpr._init_num_threads())
+                 if 'sparc' in platform.machine():
+                     self.assertEqual(1, numexpr._init_num_threads())
+                 else:
+                     self.assertEqual(3, numexpr._init_num_threads())
 
     def test_omp_num_threads(self):
         with _environment('OMP_NUM_THREADS', '5'):
-            self.assertEqual(5, numexpr._init_num_threads())
+            if 'sparc' in platform.machine():
+                self.assertEqual(1, numexpr._init_num_threads())
+            else:
+                self.assertEqual(5, numexpr._init_num_threads())
 
     def test_vml_threads_round_trip(self):
         n_threads = 3
@@ -1079,10 +1231,6 @@ def print_versions():
     from numexpr.cpuinfo import cpu
     import platform
 
-    np_version = LooseVersion(np.__version__)
-
-    if np_version < minimum_numpy_version:
-        print('*Warning*: NumPy version is lower than recommended: %s < %s' % (np_version, minimum_numpy_version))
     print('-=' * 38)
     print('Numexpr version:   %s' % numexpr.__version__)
     print('NumPy version:     %s' % np.__version__)
@@ -1111,7 +1259,6 @@ def test(verbosity=1):
     """
     Run all the tests in the test suite.
     """
-
     print_versions()
     # For some reason, NumPy issues all kinds of warnings when using Python3.
     # Ignoring them in tests should be ok, as all results are checked out.
